from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import os
import cv2
import sys
import numpy as np
sys.path.append('./lib')

from centernet.lib.opts import opts
from centernet.lib.detectors.detector_factory import detector_factory


class CDetector:
    def __init__(self, model, res_path):
        opt = opts().init()
        opt.load_model = model

        opt.vis_thresh = 0.1
        opt.debug = 0  # show

        opt.flip_test = True
        opt.nms = True  # must be False for embedding

        Detector = detector_factory[opt.task]
        self.detector = Detector(opt)

        self.cls_id = []

        self.res_path = res_path

    def init(self):  # BGR [x y w h]
        if len(self.cls_id) == 0:
            self.cls_id = list(range(1, 81))

    def detect(self, img):  # BGR
        out = self.detector.run(img)
        ret = out['results']

        tmp_box = []
        tmp_score = []
        if len(self.cls_id) > 0:
            for k in self.cls_id:
                if len(ret[k]) > 0:
                    boxes = ret[k][:, :4]
                    score = ret[k][:, 4]
                    boxes[:, 0::2] = np.clip(boxes[:, 0::2], 0, img.shape[1]-1)
                    boxes[:, 1::2] = np.clip(boxes[:, 1::2], 0, img.shape[0]-1)
                    # boxes[:, 2:] = boxes[:, 2:] - boxes[:, :2] + 1  # [x y w h]
                    tmp_box.append(boxes)
                    tmp_score.append(score)
        else:
            for k in ret.keys():
                if len(ret[k]) > 0:
                    boxes = ret[k][:, :4]
                    score = ret[k][:, 4]
                    boxes[:, 0::2] = np.clip(boxes[:, 0::2], 0, img.shape[1] - 1)
                    boxes[:, 1::2] = np.clip(boxes[:, 1::2], 0, img.shape[0] - 1)
                    # boxes[:, 2:] = boxes[:, 2:] - boxes[:, :2] + 1  # [x y w h]
                    tmp_box.append(boxes)
                    tmp_score.append(score)

        if len(tmp_box) > 0:
            tmp_box = np.concatenate(tmp_box, axis=0)
            tmp_score = np.concatenate(tmp_score, axis=0)
        else:
            tmp_box = np.array([]).reshape(-1, 4)
            tmp_score = np.array([])
        return tmp_box, tmp_score

    def detect_read(self, f_idx, vid_name):  # BGR
        cdet_box = np.load(os.path.join(self.res_path, '{}/{:0>4d}.npy'.format(vid_name, f_idx)))

        tmp_box = cdet_box[:, :4]
        tmp_score = cdet_box[:, 4]

        return tmp_box, tmp_score


    def overlap_ratio(self, r1, r2, mode='xyxy'):
        """
        computing IoU
        :param r1: (y0, x0, y1, x1), which reflects
                (top, left, bottom, right)
        :param r2: (y0, x0, y1, x1)
        :return: scala value of IoU
        """
        rec1 = r1.copy().reshape(-1, 4)
        rec2 = r2.copy().reshape(-1, 4)

        if mode == 'xywh':
            # to [x1 y1 x2 y2]
            rec1[:, 2:] = rec1[:, 2:] + rec1[:, :2]
            rec2[:, 2:] = rec2[:, 2:] + rec2[:, :2]

        # computing area of each rectangles
        s_rec1 = (rec1[:, 2] - rec1[:, 0]) * (rec1[:, 3] - rec1[:, 1])
        s_rec2 = (rec2[:, 2] - rec2[:, 0]) * (rec2[:, 3] - rec2[:, 1])

        # find the each edge of intersect rectangle
        left_line = np.maximum(rec1[:, 0], rec2[:, 0])
        right_line = np.minimum(rec1[:, 2], rec2[:, 2])
        top_line = np.maximum(rec1[:, 1], rec2[:, 1])
        bottom_line = np.minimum(rec1[:, 3], rec2[:, 3])

        intersect = (right_line - left_line) * (bottom_line - top_line)
        zero_idx = np.logical_or(left_line >= right_line, top_line >= bottom_line)
        intersect[zero_idx] = 0

        iou = intersect / (s_rec1 + s_rec2 - intersect)

        iou[zero_idx] = 0

        return iou

    def get_subwindow(self, im, pos, model_sz, original_sz, avg_chans):
        """
        args:
            im: bgr based image
            pos: center position
            model_sz: exemplar size
            s_z: original size
            avg_chans: channel average
        """
        if isinstance(pos, float):
            pos = [pos, pos]
        sz = original_sz
        im_sz = im.shape
        c = (original_sz + 1) / 2
        # context_xmin = round(pos[0] - c) # py2 and py3 round
        context_xmin = np.floor(pos[0] - c + 0.5)
        context_xmax = context_xmin + sz - 1
        # context_ymin = round(pos[1] - c)
        context_ymin = np.floor(pos[1] - c + 0.5)
        context_ymax = context_ymin + sz - 1
        left_pad = int(max(0., -context_xmin))
        top_pad = int(max(0., -context_ymin))
        right_pad = int(max(0., context_xmax - im_sz[1] + 1))
        bottom_pad = int(max(0., context_ymax - im_sz[0] + 1))

        context_xmin = context_xmin + left_pad
        context_xmax = context_xmax + left_pad
        context_ymin = context_ymin + top_pad
        context_ymax = context_ymax + top_pad

        r, c, k = im.shape
        if any([top_pad, bottom_pad, left_pad, right_pad]):
            size = (r + top_pad + bottom_pad, c + left_pad + right_pad, k)
            te_im = np.zeros(size, np.uint8)
            te_im[top_pad:top_pad + r, left_pad:left_pad + c, :] = im
            if top_pad:
                te_im[0:top_pad, left_pad:left_pad + c, :] = avg_chans
            if bottom_pad:
                te_im[r + top_pad:, left_pad:left_pad + c, :] = avg_chans
            if left_pad:
                te_im[:, 0:left_pad, :] = avg_chans
            if right_pad:
                te_im[:, c + left_pad:, :] = avg_chans
            im_patch = te_im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]
        else:
            im_patch = im[int(context_ymin):int(context_ymax + 1),
                       int(context_xmin):int(context_xmax + 1), :]

        if not np.array_equal(model_sz, original_sz):
            im_patch = cv2.resize(im_patch, (model_sz, model_sz))

        return im_patch




if __name__ == '__main__':
    vids = os.listdir('/data1/Dataset/VOT/RGBD19')
    vids.sort()
    vids.remove('list.txt')
    vids.remove('VOT2019-RGBD.json')

    iou_avg = []
    for title in vids[0:]:
        # title = vids[0]
        path = '/home/space/data/VOT2019RGBD/{}'.format(title)
        num = len(os.listdir(os.path.join(path, 'color')))

        cdet = CDetector('/home/space/Documents/experiment/VOT20RGBD/src/centernet/models/ctdet_coco_resdcn18.pth')
        for frame_i in range(0, num):
            im = cv2.imread(os.path.join(path, 'color/{:0>8d}.jpg'.format(frame_i+1)))
            x, y, w, h = np.loadtxt(os.path.join(path, 'groundtruth.txt'), delimiter=',')[frame_i]
            if frame_i == 0:
                cdet.init()

            boxes, scores = cdet.detect(im)  # [x y x y]
            boxes = np.concatenate((boxes.reshape(-1, 4), scores.reshape(-1, 1)), axis=1)

            base_results_path = os.path.join(
                '/home/space/Documents/experiment/VOT20RGBD/src/centernet/VOT2019-RGBD/{}'.format(title))
            results_path = '{}/{:0>4d}.npy'.format(base_results_path, frame_i+1)

            if not os.path.exists(base_results_path):
                os.makedirs(base_results_path)
            np.save(results_path, boxes)
            print(title, frame_i)

            # if boxes.size > 0:
            #     for b in boxes:
            #         b = b.astype(int)
            #         cv2.rectangle(im, (b[0], b[1]), (b[2], b[3]), (255, 0, 0), 2)
            #
            # cv2.imshow('', im)
            # cv2.waitKey(1)
